/* written by John Doug Reynolds, July 2002 */

#include "rgsspectrum.h"
#include <ShapeScanner.h>
#include <rgsPixelsLib.h>
#include <DalVec.h>
#include <CalServer.h>
#include <ASCregion.h>
#include <cstdio>
#include <algorithm>

namespace {
using namespace rgsPixelsLib;

struct Statistics {
	float backscal, frac_exp;
	unsigned counts;
};

class Accumulator {
public:
	list<Span> spans;
	double exposure;
	unsigned counts;

	Accumulator() : exposure(0), counts(0) {}
	void select( const ExposureMap&, int column, int offset );
};

class SpectralBin {
public:
	struct Accumulator foreground, background;
	list<Span> bananaSpans;
	float fullExposure;
	bool dubious;

	SpectralBin() : fullExposure(0), dubious(false) {}
	void fgSpans( const list<Span> &s ) { foreground.spans = s; }
	void bgSpans( const list<Span> &s ) { background.spans = s; }
	void phSpans( const list<Span> &s ) { bananaSpans = s; }
};

class Machine {
public:
	Machine( const GridSpan &beta, const GridSpan &xdsp,
			bool modeIsHTR, int source, int order, int rebin );

	void loadRegions( struct DataSet *sources, bool withfg, bool withbg );
	void loadExposure( struct Array *emap, int edgechannels );
	void loadEvents( const Table *events );

	void out( const string &dsname,
			const list<const Attribute*>&, const string &hduclas2,
			bool withfracexp, float exposed, int badquality, string spectrumbinning
	);
	//AI: Wavelength scale. Start
	void reproject(DalVecI &channel,DalVecJ &val,DalVecE &backscale, DalVecE &areascale, DalVecI &quality,
			DalVecE &statErr, DalVecE &fracExp, bool bkgSpec, bool netSpectrum, bool withfracexp );
	void setMLambdaChannel(const GridSpan &mlambda);
	//AI: Wavelength scale. End
private:
	typedef void (SpectralBin::*spaninit_f)( const list<Span>& );

	vector<SpectralBin> _accum;
	//AI: Wavelength scale. Start
	vector<SpectralBin> _lambdaAccum;
	//AI: Wavelength scale. End
	struct BackgroundDataServer* _caldsBG;
	struct HtrBackgroundIntensity _bkgModel;
	struct GridSpan _grid, _betaGrid, _xdspGrid;
	//AI: Wavelength scale. Start
	//mlambda grid
	struct GridSpan _mlGrid, _mlambdaGrid;
	double _offaxis;
	double  _alpha0;
	vector<int> _mlambdaFgdAccum, _mlambdaBkgAccum;
	bool _lambdaSpectrumBinning;
	//AI: Wavelength scale. End
	double _maxExposure, _lineSep, _rgsAlpha;
	int _source, _order, _rebin;
	bool _modeIsHTR;

	void scan( struct Table*, spaninit_f, bool spatial );
	struct Statistics reduce( const SpectralBin&, bool foreground ) const;
	float bgIntegral( const list<Span>& ) const;

	void out( struct DataSet*,
			const list<const Attribute*>&, const string &hduclas2,
			bool withfracexp, float exposed, int badquality,string spectrumbinning
	);

	static double quotient( double x, double y ) { return y ? x/y : 0; }
	static bool contains( const list<Span>&, short );
	static unsigned length( const list<Span>& );
};
}

void rgsspectrum()
{
	bool bgcorrect = false;
	string fgsetname, bgsetname;

	string dsname	= stringParameter("evlist");
	string sdefsetname	= stringParameter("srclist");
	bool withfgset	= booleanParameter("withspectrum");
	if ( withfgset ) {
		bgcorrect		= booleanParameter("bkgcorrect");
		fgsetname		= stringParameter("spectrumset");
	}
	bool withbgset	= booleanParameter("withbkgset");
	if ( withbgset )
		bgsetname		= stringParameter("bkgset");
	int source		= intParameter("source");
	int order		= intParameter("order");
	int rebin		= intParameter("rebin");
	int edgechannels	= intParameter("edgechannels");
	int badquality	= intParameter("badquality");
	float exposed	= realParameter("exposed");
	bool withfracexp	= booleanParameter("withfracexp");

	//AI: Wavelength scale. Start
	string spectrumbinning = stringParameter("spectrumbinning");
	//AI: Wavelength scale. End

	if ( !withfgset && !withbgset ) {
		Sas::warning("NothingRequested","both outputs have been disabled");
		return;
	}

	struct DataSet *dataset = dataSetServer->open( dsname, Dal::Read );

	string instrument = dataset->attribute("INSTRUME")->asString();
	string   datamode = dataset->attribute("DATAMODE")->asString();

	bool modeIsHTR = ( Xmm::toDataMode(datamode) == Xmm::HIGH_TIME_RESOLUTION );

	calServer.setInstrument( Xmm::toInstrument(instrument) );
	if ( modeIsHTR )
		calServer.state()->mode()->set( Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD );

	const Table *etab = dataset->table("EVENTS");
	if ( source == 0 ) source = etab->attribute("SOURCEID")->asInt();

	char scratch[32];
	string::size_type suffix = dsname.rfind('.');
	if ( suffix == string::npos ) suffix = dsname.size();
	/* AI gcc4 porting
	 * error: suggest explicit braces to avoid ambiguous ‘else’
	 */
	if ( sdefsetname == "" )
	{
		if ( suffix >= 10 )
			(sdefsetname=dsname).replace( suffix-10, 6, "SRCLI_" );
		else
			sdefsetname = string("srclist") + dsname.substr(suffix);
	}

	if ( withfgset && fgsetname == "" )
	{
		if ( suffix >= 10 ) {
			sprintf( scratch, "SRSPEC%d%03d", order, source );
			(fgsetname=dsname).replace( suffix-10, 10, scratch );
		}
		else {
			sprintf( scratch, "src%do%d.pha", source, order );
			fgsetname = scratch;
		}
	}

	if ( withbgset && bgsetname == "" )
	{
		if ( suffix >= 10 ) {
			sprintf( scratch, "BGSPEC%d%03d", order, source );
			(bgsetname=dsname).replace( suffix-10, 10, scratch );
		}
		else {
			sprintf( scratch, "bkg%do%d.pha", source, order );
			bgsetname = scratch;
		}
	}

	//AI: Wavelength scale. Start
	const struct Column* channelCol = 0;
	channelCol = etab->column("BETA_CHANNEL");

	struct Machine m( gridspan( channelCol ),
			gridspan( etab->column("XDSP_CHANNEL") ),
			modeIsHTR, source, order, rebin
	);

	if (spectrumbinning == "lambda")
	{
		if (!etab->hasColumn("MLAMBDA_CHANNEL"))
				Sas::error("LambdaSpectrum","MLAMBDA_CHANNEL column needed.");
		const struct Column* mlambdaChannelCol = 0;
		mlambdaChannelCol = etab->column("MLAMBDA_CHANNEL");
		m.setMLambdaChannel(gridspan(mlambdaChannelCol));
	}
	//AI: Wavelength scale. End

	struct DataSet *sdefset = dataSetServer->open( sdefsetname, Dal::Read );
	m.loadRegions( sdefset, withfgset, (withfgset&&bgcorrect)||withbgset );
	dataSetServer->close( sdefset );

	for ( unsigned ccd = 1; ccd <= 9; ++ccd )
		for ( unsigned node = 0; node < 2; ++node ) {
			char extname[9];
			sprintf( extname, "EXPMAP%u%u", node, ccd );
			if ( dataset->hasBlock(extname) )
				m.loadExposure( dataset->array(extname), edgechannels );
		}

	m.loadEvents( etab );

	list<const Attribute*> attrib;
	attrib.push_back( dataset->attribute("TELESCOP") );
	attrib.push_back( dataset->attribute("INSTRUME") );
	attrib.push_back( dataset->attribute("FILTER") );
	attrib.push_back( dataset->attribute("OBJECT") );
	attrib.push_back( dataset->attribute("OBS_ID") );
	attrib.push_back( dataset->attribute("EXP_ID") );
	attrib.push_back( dataset->attribute("DATE-OBS") );
	attrib.push_back( dataset->attribute("DATE-END") );
	attrib.push_back( dataset->attribute("EQUINOX") );
	attrib.push_back( dataset->attribute("RADECSYS") );
	attrib.push_back( etab->attribute("RA_OBJ") );
	attrib.push_back( etab->attribute("DEC_OBJ") );

	if ( dataset->hasAttribute("EXPIDSTR") )
		attrib.push_back( dataset->attribute("EXPIDSTR") );

	//AI: Wavelength scale. Start
	// passing spectrumbinning
	if ( withfgset )
		m.out( fgsetname, attrib, bgcorrect ? "NET" : "TOTAL",
				withfracexp, exposed, badquality,spectrumbinning );

	if ( withbgset )
		m.out( bgsetname, attrib, "BKG", withfracexp, exposed, badquality,spectrumbinning );
	//AI: Wavelength scale. End

	dataSetServer->close( dataset );
}

/* --- Machine methods ----------------------------------------------------- */
void Machine::setMLambdaChannel(const GridSpan &mlambda)
{
	//AI: Wavelength scale. Start
	//Do the same as _grid
	// Initialize lambda accumulator
	_mlambdaGrid = mlambda;
	size_t mlnbins = (size_t) ceil( mlambda.length() / (float)_rebin );
	_mlGrid = GridSpan( Grid( _rebin*mlambda.stride, mlambda[(.5-mlambda.min)*(_rebin-1)] ),
			Span(mlambda.min).length(mlnbins) );
	_lambdaAccum.resize(mlnbins);

	_lambdaSpectrumBinning = true;
	//AI: Wavelength scale. End
}

unsigned Machine::length( const list<Span> &spans )
{
	unsigned sum = 0;
	list<Span>::const_iterator itr;
	for ( itr = spans.begin(); itr != spans.end(); ++itr ) sum += itr->length();
	return sum;
}

bool Machine::contains( const list<Span> &spans, short p )
{
	list<Span>::const_iterator itr;
	for ( itr = spans.begin(); itr != spans.end(); ++itr )
		if ( p <= itr->max ) return (p >= itr->min);
	return false;
}
//AI: Wavelength scale. Start
Machine::Machine( const GridSpan &beta, const GridSpan &xdsp,
		bool modeIsHTR, int source, int order, int rebin )
:
	_bkgModel(0), // no default constructor is provided, hence this kludge
	_betaGrid(beta), _xdspGrid(xdsp), _maxExposure(0),
	_source(source), _order(order), _rebin(rebin), _modeIsHTR(modeIsHTR)
	{
	_caldsBG = calServer.getAtom( _caldsBG );

	// rebin: minval remains same, maxval grows as necessary
	size_t nbins = (size_t) ceil( beta.length() / (float)rebin );
	_grid = GridSpan( Grid( rebin*beta.stride, beta[(.5-beta.min)*(rebin-1)] ),
			Span(beta.min).length(nbins) );
	_accum.resize(nbins);

	//AI: Wavelength scale.
	//Set false by default
	_lambdaSpectrumBinning = false;
	}

void Machine::scan( struct Table *table, spaninit_f initf, bool spatial )
{
	struct InASCregionTable ascReg(table);
	struct GeoRegion geoReg;
	ascReg >> geoReg;

	struct RegionScanner scanner( geoReg, _grid,
			spatial ? static_cast<const Grid &>(_xdspGrid) : Grid() );
	struct Bounds clip(_grid,spatial ? (Span&)_xdspGrid : scanner.bounds().row);
	scanner.reset(&clip);

	int b;
	list<Span> spans;
	while ( scanner.next(b,spans) ) (_accum[b - _grid.min].*initf)(spans);
}

void Machine::loadRegions( struct DataSet *dataset, bool withfg, bool withbg )
{
	// initialize _lineSep and _rgsAlpha

	unsigned i = 0;
	const Table *table = dataset->table("SRCLIST");
	const DalVecI index( table->column("INDEX") );
	while ( i < table->rows() && index[i] != _source ) ++i;
	if ( i == table->rows() )
		errstr << "source INDEX=" << _source << " not found in "
		<< table->qualifiedName() << Sas::error("SourceNotFound");

	/*
	 *  AI gcc4 porting
	 *   error: ‘misc’ may be used uninitialized in this function
	 *   => misc = calServer.getAtom(misc);
	 */
	struct MiscDataServer *misc = 0;
	misc = calServer.getAtom(misc);

	//AI: Wavelegnth scale. Start
	_alpha0 = (misc->get("INCIDENCE_ANGLE")
			+ misc->get("LINCOORD:LINCOORD%G_RY")) * M_PI/180;
	//AI: Wavelegnth scale. End

	_lineSep = misc->get("GRAT_LINE_DENS");
	_rgsAlpha = (misc->get("LINCOORD:LINCOORD%G_RY")
			+ misc->get("INCIDENCE_ANGLE")
			+ (DalVecE(table->column("DELTA_DISP"))[i]/60)
	) * M_PI/180;

	// initialize the accumulator spans

	char fgSpatialReg[64], bgSpatialReg[64], fgOrderReg[64], bgOrderReg[64];

	string instrument = dataset->attribute("INSTRUME")->asString();
	const char *instrCStr = instrument.c_str();
	sprintf( bgSpatialReg, "%s_BACKGROUND",      instrCStr );
	sprintf( fgSpatialReg, "%s_SRC%d_SPATIAL",   instrCStr, _source );
	sprintf(   fgOrderReg, "%s_SRC%d_ORDER_%d",  instrCStr, _source, _order );
	sprintf(   bgOrderReg, "%s_SRC%d_ORD%d_BKG", instrCStr, _source, _order );

	if ( _modeIsHTR ) {
		if ( withfg )
			scan( dataset->table(fgOrderReg), &SpectralBin::phSpans, false );

		if ( withbg )
			scan( dataset->table(bgOrderReg), &SpectralBin::bgSpans, false );

		scan( dataset->table(fgSpatialReg), &SpectralBin::fgSpans, true );
	}
	else {
		if ( withfg )
			scan( dataset->table(fgSpatialReg), &SpectralBin::fgSpans, true );

		if ( withbg )
			scan( dataset->table(bgSpatialReg), &SpectralBin::bgSpans, true );

		scan( dataset->table(fgOrderReg), &SpectralBin::phSpans, false );
	}
}

void Machine::loadExposure( struct Array *array, int edgechannels )
{
	struct ExposureMap emap(array);
	bool valid = false;

	if ( !emap.bounds() )
		errstr << "the exposure map is null";
	else if ( !commensurate(_betaGrid,emap.beta) )
		errstr << "the exposure map dispersion channels are not "
		<< "commensurate with the EVENTS table";
	else if ( !commensurate(_xdspGrid,emap.xdsp) )
		errstr << "the exposure map cross-dispersion channels are not "
		<< "commensurate with the EVENTS table";
	else valid = true;

	if ( !valid ) {
		errstr << ": discarding " << array->qualifiedName()
		<< Sas::warning("InvalidExposureMap");
		return;
	}

	const Bounds &bnds = emap.bounds();
	int b, offset = bnds.row.min - _xdspGrid.cell( emap.xdsp[bnds.row.min] );
	double datamax = array->attribute("DATAMAX")->asReal();


	for ( int column = bnds.col.min; column <= bnds.col.max; ++column )
	{
		/* AI gcc4 porting
		 * error: suggest explicit braces to avoid ambiguous ‘else’
		 */
		if ( (b = _grid.cell( emap.beta[column] )) >= _grid.min )
		{
			if ( b <= _grid.max ) {
				struct SpectralBin &bin = _accum[ b - _grid.min ];

				if ( bin.fullExposure == 0 )
					bin.fullExposure = datamax;
				else if ( bin.fullExposure != datamax )
					bin.dubious = true;

				bin.foreground.select( emap, column, offset );
				if ( !_modeIsHTR )
					bin.background.select( emap, column, offset );
			}
			else break;
		}
	}

	int i, end;
	if ( array->attribute("CCDNODE")->asInt() ) {
		i = _grid.cell(emap.beta[bnds.col.min]) - _grid.min;
		end = i + edgechannels;
	}
	else {
		end = 1 + _grid.cell(emap.beta[bnds.col.max]) - _grid.min;
		i = end - edgechannels;
	}
	maximize( i, 0 );
	minimize( end, (int)_accum.size() );
	while ( i < end ) _accum[i++].dubious = true;

	maximize( _maxExposure, datamax );
}

void Machine::loadEvents( const Table *table )
{

	//AI: Wavelength scale. Start
	if (_lambdaSpectrumBinning)
	{
		_offaxis = 0.0;
		if (table->hasAttribute("OFFAXIS") )
			_offaxis =  table->attribute("OFFAXIS")->asReal();

		if (!table->hasColumn("M_LAMBDA"))
			Sas::error("LambdaSpectrum","M_LAMBDA column needed.");
	}


	const DalVecE mlambda(table->column("M_LAMBDA")) ;

	//AI: Wavelength scale. End
	const struct Column* channelCol = 0;
	channelCol = table->column("BETA_CHANNEL");

	const DalVecI beta( channelCol );
	const DalVecI xdsp( table->column("XDSP_CHANNEL") );
	const DalVecI calpi( table->column("PI") );
	for ( unsigned i = 0; i < table->rows(); ++i ) {

		int b = (beta[i] - _grid.min) / _rebin;
		if ( _grid.contains(b+_grid.min) ) {
			struct SpectralBin &bin = _accum[b];
			int index = (_mlambdaGrid.cell(mlambda[i])- _grid.min) /_rebin;
			//AI: Wavelength scale. Start
			//struct SpectralBin &lambdaBin =
			//	_lambdaAccum[index];
			struct SpectralBin *lambdaBin = 0;
			if (_lambdaSpectrumBinning && (index >=0))
			{
				lambdaBin =
					&(_lambdaAccum[_mlambdaGrid.cell(mlambda[i])- _grid.min]);
			}
			//AI: Wavelength scale. End

			short c = xdsp[i], e = calpi[i];

			// calculate the corresponding mlambda value for the current event
			// and store it in the corresponding mlambda bin
			if ( _modeIsHTR ) {
				if ( contains(bin.foreground.spans,c) ) {
					if ( contains(bin.bananaSpans,e) )
					{
						++bin.foreground.counts;
						if (_lambdaSpectrumBinning  && (index >=0))
						  ++(*lambdaBin).foreground.counts;       //AI: Wavelength scale.
					}
					if ( contains(bin.background.spans,e) )
					{
						++bin.background.counts;
						if (_lambdaSpectrumBinning  && (index >=0))
						  ++(*lambdaBin).background.counts;       //AI: Wavelength scale. Start
					}
				}
			}
			else
				if ( contains(bin.bananaSpans,e) ) {
					if ( contains(bin.foreground.spans,c) )
					  {
						++bin.foreground.counts;
						if (_lambdaSpectrumBinning && (index >=0) )
						  ++(*lambdaBin).foreground.counts;    //AI: Wavelength scale. Start
					}
					if ( contains(bin.background.spans,c) )
					{
						++bin.background.counts;
						if (_lambdaSpectrumBinning  && (index >=0))
						  ++(*lambdaBin).background.counts;   //AI: Wavelength scale. Start
					}
				}
		}
	}
}

void
Machine::out( const string &dsname,
		const list<const Attribute*> &attrib, const string &hduclas2,
		bool withfracexp, float exposed, int badquality,string spectrumbinning )
{
	if ( _maxExposure == 0 )
		Sas::error("ZeroExposure","no exposure found in any node");

	struct DataSet *dataset = dataSetServer->open( dsname, Dal::Create );

	list<const Attribute*>::const_iterator itr;
	for ( itr = attrib.begin(); itr != attrib.end(); ++itr )
		dataset->addAttribute(*itr);

	out( dataset, attrib, hduclas2, withfracexp, exposed, badquality, spectrumbinning );

	dataSetServer->close( dataset );
}

float Machine::bgIntegral( const list<Span> &spans ) const
{
	float sum = 0;
	list<Span>::const_iterator itr;
	for ( itr = spans.begin(); itr != spans.end(); ++itr )
		sum += _bkgModel.accumulatedIntensity( itr->min - .5, itr->max + .5 );
	return sum;
}

struct Statistics Machine::reduce( const SpectralBin &bin, bool fgStats ) const
{
	struct Statistics stats;
	if ( _modeIsHTR ) {
		if ( fgStats ) {
			stats.counts = bin.foreground.counts;
			stats.backscal = bgIntegral( bin.bananaSpans );
		}
		else {
			stats.counts = bin.background.counts;
			stats.backscal = bgIntegral( bin.background.spans );
		}
		unsigned pixels = _rebin * length( bin.foreground.spans );
		stats.frac_exp =
			quotient( bin.foreground.exposure, pixels * bin.fullExposure );
	}
	else {
		const Accumulator &accum = fgStats ? bin.foreground : bin.background;
		stats.counts = accum.counts;
		stats.backscal = quotient( accum.exposure, bin.fullExposure );
		unsigned pixels = _rebin * length( accum.spans );
		stats.frac_exp = quotient( accum.exposure, pixels * bin.fullExposure );
	}
	return stats;
}

void
Machine::out( struct DataSet *dataset,
		const list<const Attribute*> &attrib, const string &hduclas2,
		bool withfracexp, float exposed, int badquality,string spectrumbinning )
{

	bool netSpectrum = (hduclas2 == "NET");
	bool bkgSpectrum = (hduclas2 == "BKG");
	//AI: Wavelength scale. Start
	struct Table *table;
	if (spectrumbinning == "lambda")
		table = dataset->addTable( "SPECTRUM", _lambdaAccum.size() );
	else
		table = dataset->addTable( "SPECTRUM", _accum.size() );
	//AI: Wavelength scale. End
	list<const Attribute*>::const_iterator itr;
	for ( itr = attrib.begin(); itr != attrib.end(); ++itr )
		table->addAttribute(*itr);

#define ADD(k,r,u,v) *table->addAttribute(k,r,u) = v
	ADD("HDUCLASS","format conforms to OGIP standard","","OGIP");
	ADD("HDUCLAS1","OGIP Memo OGIP/92-007","","SPECTRUM");
	ADD("HDUCLAS2","what combination of source and background","",hduclas2);
	ADD("HDUCLAS3","which data representation","","COUNT");
	ADD("HDUVERS","version of format specification","","1.2.0");
	ADD("POISSERR","whether Poissonian errors may be assumed","",!netSpectrum);
	ADD("CHANTYPE","which channel calibration applies","","PI");
	//AI: Wavelength scale. Start
	if (spectrumbinning == "lambda")
		ADD("DETCHANS","total number of detector channels","",(int)_lambdaAccum.size());
	else
		ADD("DETCHANS","total number of detector channels","",(int)_accum.size());
	//AI: Wavelength scale. Start
	ADD("SYS_ERR","no systematic error specified","",0);
	ADD("GROUPING","no channel grouping specified","",0);
	ADD("CORRSCAL","correlation scale factor","",1);
	ADD("CORRFILE","associated correlation file","","none");
	ADD("RESPFILE","associated redistribution matrix","","none");
	ADD("ANCRFILE","associated ancillary response file","","none");
	ADD("BACKFILE","associated background spectrum","","none");
	ADD("SOURCEID","source extracted","SRCLIST:INDEX",_source);
	ADD("RFLORDER","reflection order extracted","",_order);
	ADD("LINE_SEP","grating line spacing","Angstroms",_lineSep);
	ADD("RGSALPHA","grating incidence angle","rad",_rgsAlpha);
	ADD("EXPOSURE","maximum effective integration time","s",_maxExposure);
	// Background corrected RGS spectra cannot be read int XSPEC 12
	// We introduce a dummy BACKSCAL keyword (S0C-SPR-6647)
	if (netSpectrum)
		ADD("BACKSCAL","background scaling factor","", 1.0);

#undef ADD

	DalVecE stat_err, backscal, frac_exp;
	DalVecI channel(   table, "CHANNEL",  "dispersion channel" );
	DalVecJ  counts(   table, "COUNTS",   "photon detection count", "counts" );
	if ( netSpectrum )
		stat_err.reset( table, "STAT_ERR", "statistical error", "counts" );
	DalVecI  quality(  table, "QUALITY",  "quality flag" );
	DalVecE areascal(  table, "AREASCAL", "area scaling factor" );
	if ( !netSpectrum )
		backscal.reset( table, "BACKSCAL", "background scaling factor" );
	if ( withfracexp )
		frac_exp.reset( table, "FRAC_EXP", "fraction of full exposure" );

	//AI: Wavelength scale. Start
	//MLAMBDA CHANNEL

	struct Column *chc = channel.column();
	if (spectrumbinning == "lambda" )
	{
		*chc->addAttribute("TCTYP","channel axis type") = "";
		*chc->addAttribute("TCUNI","channel axis units(-Angstrom*order)") = "Angstrom";
		*chc->addAttribute("TCRPX","reference channel number") = _mlGrid.min;
		*chc->addAttribute("TCRVL","center of reference channel")= _mlGrid[_mlGrid.min];
		*chc->addAttribute("TCDLT","uniform channel width") = _mlGrid.stride;
		*chc->addAttribute("TLMIN") = _mlGrid.min;
		*chc->addAttribute("TLMAX") = _mlGrid.max;
	}
	else
	{
		*chc->addAttribute("TCTYP","channel axis type") = "";
		*chc->addAttribute("TCUNI","channel axis units") = "rad";
		*chc->addAttribute("TCRPX","reference channel number") = _grid.min;
		*chc->addAttribute("TCRVL","center of reference channel")= _grid[_grid.min];
		*chc->addAttribute("TCDLT","uniform channel width") = _grid.stride;
		*chc->addAttribute("TLMIN") = _grid.min;
		*chc->addAttribute("TLMAX") = _grid.max;
	}


	//AI: Wavelength scale. End

	for ( unsigned i = 0; i < _accum.size(); ++i ) {
		const SpectralBin &bin = _accum[i];
		int chn = _grid.min + (int)i;

		channel[i] = chn;
		areascal[i] = bin.fullExposure / _maxExposure;
		quality[i] = bin.dubious ? badquality : 0;

		if ( _modeIsHTR ) _bkgModel = _caldsBG->betaSelect( _grid[chn] );

		if ( netSpectrum ) {
			struct Statistics fg = reduce(bin,true), bg = reduce(bin,false);
			float scale = quotient( fg.backscal, bg.backscal );
			counts[i] = fg.counts - int( .5 + scale*bg.counts );
			stat_err[i] = std::max(1.,static_cast<double>(std::sqrt(fg.counts + scale*scale*bg.counts )));
			if ( withfracexp )
				frac_exp[i] = fg.frac_exp;
			if ( !bg.backscal || fg.frac_exp < exposed || bg.frac_exp < exposed )
				quality[i] = badquality;
		}
		else {
			struct Statistics stats = reduce( bin, !bkgSpectrum );
			counts[i] = stats.counts;

			backscal[i] = stats.backscal;
			if ( withfracexp )
				frac_exp[i] = stats.frac_exp;
			if ( stats.frac_exp < exposed )
				quality[i] = badquality;
		}
	}
	//AI: Wavelength scale. Start
	//Reproject the beta channel into a lambda channel and distribute the counts
	//into the new lambda grid.
	if (spectrumbinning == "lambda")
		reproject(channel,counts,backscal,areascal,quality,stat_err,frac_exp,bkgSpectrum,netSpectrum,withfracexp);

}


void Machine::reproject(DalVecI &channel,DalVecJ &val, DalVecE &backscale, DalVecE &areascale, DalVecI &quality,
		DalVecE &statErr, DalVecE &fracExp, bool bkgSpec, bool netSpectrum, bool withfracexp )
{
	errstr<<"Calculating areascal, backscale in lambda grid" << endl
	<<Sas::message(Msg::App, Msg::Verbose);

	vector<int> lambdaQuality(_mlGrid.length());
	vector<double> lambdaBackscale(_mlGrid.length());
	vector<double> lambdaBkgBackscale(_mlGrid.length());
	vector<double> lambdaAreascale(_mlGrid.length());
	vector<double> lambdaStatErr(_mlGrid.length());
	vector<double> lambdaFracExp(_mlGrid.length());
	vector<int> lambdaCounts(_mlGrid.length());

	vector<double> fgBackscal(_accum.size());
	vector<double> bgBackscal(_accum.size());

	// Mark all the bins as bad quality.
	for (unsigned int i = 0 ; i < lambdaQuality.size() ; ++i)
		lambdaQuality[i] = 1;


	if ( netSpectrum ) {
		for ( unsigned i = 0; i < _accum.size(); ++i ) {
			const SpectralBin &bin = _accum[i];
			struct Statistics fg = reduce(bin,true), bg = reduce(bin,false);
			fgBackscal[i] = fg.backscal;
			bgBackscal[i] = bg.backscal;
		}
	}



	for (unsigned int j = 0 ; j < _mlGrid.length(); ++j)
	{
		int chn = j ;
		double lambdaMin = _mlGrid[chn] - (_mlGrid.stride/2.);
		double lambdaMax = _mlGrid[chn] + (_mlGrid.stride/2.);
		//cout << "lambdaMin " << lambdaMin << " lambdaMax "<< lambdaMax
		//<< " lambdaMax - lambdaMin " << lambdaMax - lambdaMin<< endl;
		//cout << " _grid.stride "<< _grid.stride << endl;
		double cosBetaMin = cos(_alpha0 + _offaxis) + (-1.*lambdaMin)/(_lineSep);
		double cosBetaMax = cos(_alpha0 + _offaxis) + (-1.*lambdaMax)/(_lineSep);

		if (cosBetaMin > 1.0)
			cosBetaMin = 0.0;
		if (cosBetaMax > 1.0)
			cosBetaMax = 0.0;

		double acosBetaMin = acos(cosBetaMin);
		double acosBetaMax = acos(cosBetaMax);
		//cout << "betaMin  "<< acosBetaMin << " betaMax "<< acosBetaMax <<" betaWidth " << acosBetaMax-acosBetaMin << endl;
		//double ratio = (acosBetaMax - acosBetaMin)/_grid.stride;
		//cout << " ratio "<< ratio << endl;

 		int begin = _grid.cell(acosBetaMin);
		int end = _grid.cell(acosBetaMax);

		//cout << "cell bin begin " << begin << " cell bin end " << end << endl;
		//cout << "[ " << _grid[begin] - ( _grid.stride /2.) << " , " <<  _grid[begin] + ( _grid.stride /2.) << " ]" << endl;
		//cout << "[ " << _grid[end] - ( _grid.stride /2.) << " , " <<  _grid[end] + ( _grid.stride /2.) << " ]" << endl;
		if ((begin >= 0) && (end >= 0))
		{
			double tickMin = acosBetaMin;
			double tickMax = acosBetaMax;
			double counter = 0.0;
			while (begin <= end)
			{

				double binMin =  _grid[begin] - ( _grid.stride /2.);
				double binMax =  _grid[begin] + ( _grid.stride /2.);

				if (acosBetaMax <= binMax)
					tickMax = acosBetaMax;
				else
					tickMax = binMax;

				double binFraction = (tickMax - tickMin)/(binMax - binMin);
				counter += binFraction;

				//cout << "Ticks  [ " << tickMin << " , " << tickMax << " ]" <<endl;
				//cout << "Limits [ " << binMin << " , " << binMax << " ] "<< endl;
				//cout << "Fraction " << binFraction << endl;
				//cout << "fgBacscale[ " << begin <<  " ] " << fgBackscal[begin] << endl;
				if ((unsigned(end) > _accum.size()) ||
						(unsigned(begin) > _accum.size()) )
				{
					lambdaQuality[j] = 1;
					lambdaAreascale[j] = 0.0;
					lambdaBackscale[j] = 0.0;
				}

				lambdaQuality[j] = quality[begin];

				lambdaAreascale[j] += binFraction * areascale[begin];

				if ( !netSpectrum )
					lambdaBackscale[j] += binFraction * backscale[begin];
				else
				{
					lambdaBackscale[j] += binFraction * fgBackscal[begin];
					lambdaBkgBackscale[j] += binFraction * bgBackscal[begin];
				}
				if ( withfracexp )
					lambdaFracExp[j] += binFraction * fracExp[begin];

				begin++;
				tickMin = binMax;
			}

			if (unsigned(begin) > _accum.size())
			{
				lambdaQuality[j] = 1;
				lambdaAreascale[j] = 0.0;
				lambdaBackscale[j] = 0.0;
			}

			lambdaAreascale[j] /= counter;
			//cout << "lambdaAreascale[ " << j << " ] " << lambdaAreascale[j] << endl;
		}
		else
		{
			lambdaQuality[j] = 1;
			lambdaAreascale[j] = 0.0;
			lambdaBackscale[j] = 0.0;
		}


	}

	if ( netSpectrum ) {
		for (unsigned int j = 0 ; j < _lambdaAccum.size(); ++j)
		{
			float scale = quotient(lambdaBackscale[j], lambdaBkgBackscale[j]);
			lambdaCounts[j] = _lambdaAccum[j].foreground.counts - int( .5 + scale*_lambdaAccum[j].background.counts );
			lambdaStatErr[j] = std::max(1.,static_cast<double>(std::sqrt(_lambdaAccum[j].foreground.counts + scale*scale*_lambdaAccum[j].background.counts)));
		}
	}


	for (unsigned int j = 0 ; j < _lambdaAccum.size() ; ++j)
	{
		int chn = _grid.min + (int)j;
		channel[j] = chn;
		if ( netSpectrum )
			val[j] =  lambdaCounts[j];
		else if ( !bkgSpec )
		{
			val[j] =_lambdaAccum[j].foreground.counts;
		}
		else
			val[j] =_lambdaAccum[j].background.counts;

		if ( !netSpectrum )
			backscale[j] = lambdaBackscale[j];

		areascale[j] = lambdaAreascale[j];
		quality[j] = lambdaQuality[j];

		if ( netSpectrum )
			statErr[j] = lambdaStatErr[j];

		if ( withfracexp )
			fracExp[j] = lambdaFracExp[j];
	}

}





/* --- Accumulator methods ------------------------------------------------- */

void Accumulator::select( const ExposureMap &emap, int column, int offset )
{
	list<Span>::const_iterator sItr;
	for ( sItr = spans.begin(); sItr != spans.end(); ++sItr ) {
		struct Span span( offset + sItr->min, offset + sItr->max );
		if ( !(span &= emap.bounds().row) ) continue;

		long itr = emap.id( column, span.min - 1 );
		long end = emap.id( column, span.max );

		while ( itr != end )
			exposure += emap[ emap.incr(itr,Compass::North) ];
	}

}
